@@ -1,14 +1,19 @@
-import logging
 from collections import namedtuple
 from dateutil.parser import parse as dateutil_parse
 from datetime import datetime, timedelta
 import pytz
-import urllib2
 import requests
 import pandas as pd
 import zipfile
-from StringIO import StringIO
+from io import StringIO, BytesIO
+from time import sleep
+from pyiso import LOGGER
+from pytz import AmbiguousTimeError
 
+try:
+    from urllib2 import urlopen
+except ImportError:
+    from urllib.request import urlopen  # Changed from urllib2 for python3.x
 
 # named tuple for time period interval labels
 IntervalChoices = namedtuple('IntervalChoices', ['hourly', 'fivemin', 'tenmin', 'na', 'dam'])
@@ -17,16 +22,13 @@ IntervalChoices = namedtuple('IntervalChoices', ['hourly', 'fivemin', 'tenmin',
 FUEL_CHOICES = ['biogas', 'biomass', 'coal', 'geo', 'hydro',
                 'natgas', 'nonwind', 'nuclear', 'oil', 'other',
                 'refuse', 'renewable', 'smhydro', 'solar', 'solarpv',
-                'solarth', 'thermal', 'wind']
+                'solarth', 'thermal', 'wind', 'fossil', 'dual']
 
 
 class BaseClient(object):
     """
     Base class for scraper/parser clients.
     """
-    # logger
-    logger = logging.getLogger(__name__)
-
     # choices for market and frequency interval labels
     MARKET_CHOICES = IntervalChoices(hourly='RTHR', fivemin='RT5M', tenmin='RT5M', na='RT5M', dam='DAHR')
     FREQUENCY_CHOICES = IntervalChoices(hourly='1hr', fivemin='5m', tenmin='10m', na='n/a', dam='1hr')
@@ -37,6 +39,9 @@ class BaseClient(object):
     # name
     NAME = ''
 
+    # default connection timeout
+    TIMEOUT_SECONDS = 20
+
     def __init__(self):
         self.options = {}
 
@@ -48,15 +53,15 @@ class BaseClient(object):
            Available for all regions.
         :param bool yesterday: If True, get the generation mix for every time point yesterday.
            Not available for all regions.
-        :param datetime start_at: A timezone-aware datetime. The timestamp of all returned data points will be greater than or equal to this value.
+        :param datetime start_at: If the datetime is naive, it is assummed to be in the timezone of the Balancing Authority. The timestamp of all returned data points will be greater than or equal to this value.
            If using, must provide both ``start_at`` and ``end_at`` parameters.
            Not available for all regions.
-        :param datetime end_at: A timezone-aware datetime. The timestamp of all returned data points will be less than or equal to this value.
+        :param datetime end_at: If the datetime is naive, it is assummed to be in the timezone of the Balancing Authority. The timestamp of all returned data points will be less than or equal to this value.
            If using, must provide both ``start_at`` and ``end_at`` parameters.
            Not available for all regions.
         :return: List of dicts, each with keys ``[ba_name, timestamp, freq, market, fuel_name, gen_MW]``.
            Timestamps are in UTC.
-        :rtype: list 
+        :rtype: list
 
         """
         raise NotImplementedError('Derived classes must implement the get_generation method.')
@@ -69,15 +74,15 @@ class BaseClient(object):
            Available for all regions.
         :param bool yesterday: If True, get the load for every time point yesterday.
            Not available for all regions.
-        :param datetime start_at: A timezone-aware datetime. The timestamp of all returned data points will be greater than or equal to this value.
+        :param datetime start_at: If the datetime is naive, it is assummed to be in the timezone of the Balancing Authority. The timestamp of all returned data points will be greater than or equal to this value.
            If using, must provide both ``start_at`` and ``end_at`` parameters.
            Not available for all regions.
-        :param datetime end_at: A timezone-aware datetime. The timestamp of all returned data points will be less than or equal to this value.
+        :param datetime end_at: If the datetime is naive, it is assummed to be in the timezone of the Balancing Authority. The timestamp of all returned data points will be less than or equal to this value.
            If using, must provide both ``start_at`` and ``end_at`` parameters.
            Not available for all regions.
         :return: List of dicts, each with keys ``[ba_name, timestamp, freq, market, load_MW]``.
            Timestamps are in UTC.
-        :rtype: list 
+        :rtype: list
 
         """
         raise NotImplementedError('Derived classes must implement the get_load method.')
@@ -91,15 +96,15 @@ class BaseClient(object):
            Available for all regions.
         :param bool yesterday: If True, get the trade for every time point yesterday.
            Not available for all regions.
-        :param datetime start_at: A timezone-aware datetime. The timestamp of all returned data points will be greater than or equal to this value.
+        :param datetime start_at: If the datetime is naive, it is assummed to be in the timezone of the Balancing Authority. The timestamp of all returned data points will be greater than or equal to this value.
            If using, must provide both ``start_at`` and ``end_at`` parameters.
            Not available for all regions.
-        :param datetime end_at: A timezone-aware datetime. The timestamp of all returned data points will be less than or equal to this value.
+        :param datetime end_at: If the datetime is naive, it is assummed to be in the timezone of the Balancing Authority. The timestamp of all returned data points will be less than or equal to this value.
            If using, must provide both ``start_at`` and ``end_at`` parameters.
            Not available for all regions.
         :return: List of dicts, each with keys ``[ba_name, timestamp, freq, market, net_exp_MW]``.
            Timestamps are in UTC.
-        :rtype: list 
+        :rtype: list
 
         """
         raise NotImplementedError('Derived classes must implement the get_trade method.')
@@ -146,7 +151,6 @@ class BaseClient(object):
             self.options['sliceable'] = False
             self.options['forecast'] = False
 
-
     def utcify(self, local_ts_str, tz_name=None, is_dst=None):
         """
         Convert a datetime or datetime string to UTC.
@@ -168,16 +172,16 @@ class BaseClient(object):
         # parse
         try:
             local_ts = dateutil_parse(local_ts_str)
-        except AttributeError: # already parsed
+        except AttributeError:  # already parsed
             local_ts = local_ts_str
 
         # localize
-        if local_ts.tzinfo is None: # unaware
+        if local_ts.tzinfo is None:  # unaware
             if is_dst is None:
                 aware_local_ts = tz.localize(local_ts)
             else:
                 aware_local_ts = tz.localize(local_ts, is_dst=is_dst)
-        else: # already aware
+        else:  # already aware
             aware_local_ts = local_ts
 
         # convert to utc
@@ -199,11 +203,11 @@ class BaseClient(object):
         return cleaned_vals
 
     def fetch_xls(self, url):
-        socket = urllib2.urlopen(url)
+        socket = urlopen(url)
         xd = pd.ExcelFile(socket)
         return xd
 
-    def request(self, url, mode='get', **kwargs):
+    def request(self, url, mode='get', retry_sec=5, retries_remaining=5, **kwargs):
         """
         Get or post to a URL with the provided kwargs.
         Returns the response, or None if an error was encountered.
@@ -222,44 +226,74 @@ class BaseClient(object):
             session = self.session
 
         # carry out request
-       # try:
-        response = getattr(session, mode)(url, verify=False, **kwargs)
+        try:
+            response = getattr(session, mode)(url, verify=False,
+                                              timeout=self.TIMEOUT_SECONDS,
+                                              **kwargs)
         # except requests.exceptions.ChunkedEncodingError as e:
         #     # JSON incomplete or not found
         #     msg = '%s: chunked encoding error for %s, %s:\n%s' % (self.NAME, url, kwargs, e)
-        #     self.logger.error(msg)
-        #     return None
-        # except requests.exceptions.ConnectionError as e:
-        #     # eg max retries exceeded
-        #     msg = '%s: connection error for %s, %s:\n%s' % (self.NAME, url, kwargs, e)
-        #     self.logger.error(msg)
+        #     LOGGER.error(msg)
         #     return None
+        except (requests.exceptions.ConnectionError, requests.exceptions.Timeout) as e:
+            # eg max retries exceeded
+            msg = '%s: connection error for %s, %s:\n%s' % (self.NAME, url, kwargs, e)
+            LOGGER.error(msg)
+            return None
         # except requests.exceptions.RequestException:
         #     msg = '%s: request exception for %s, %s:\n%s' % (self.NAME, url, kwargs, e)
-        #     self.logger.error(msg)
+        #     LOGGER.error(msg)
         #     return None
 
         if response.status_code == 200:
-            self.logger.debug('%s: request success for %s, %s' % (self.NAME, url, kwargs))
+            # success
+            LOGGER.debug('%s: request success for %s, %s with cache hit %s' % (self.NAME, url, kwargs, getattr(response, 'from_cache', None)))
+
+        elif response.status_code == 429:
+            if retries_remaining > 0:
+                # retry on throttle
+                LOGGER.warn('%s: retrying in %d seconds (%d retries remaining), throttled for %s, %s' % (self.NAME, retry_sec, retries_remaining, url, kwargs))
+                sleep(retry_sec)
+                retries_remaining -= 1
+                return self.request(url, mode=mode,
+                                    retry_sec=retry_sec*2, retries_remaining=retries_remaining,
+                                    **kwargs)
+            else:
+                # exhausted retries
+                LOGGER.warn('%s: exhausted retries for %s, %s' % (self.NAME, url, kwargs))
+                return None
+
         else:
-            self.logger.error('%s: request failure with code %s for %s, %s' % (self.NAME, response.status_code, url, kwargs))
+            # non-throttle error
+            LOGGER.error('%s: request failure with code %s for %s, %s' % (self.NAME, response.status_code, url, kwargs))
 
         return response
 
     def unzip(self, content):
         """
         Unzip encoded data.
-        Returns the unzipped content, or None if an error was encountered.
+        Returns the unzipped content as an array of strings, each representing one file's content
+        or returns None if an error was encountered.
+        ***Previous behavior: Only returned the content from the first file***
         """
         # create zip file
         try:
-            z = zipfile.ZipFile(StringIO(content)) # have zipfile
+            filecontent = BytesIO(content)
+        except TypeError:
+            filecontent = StringIO(content)
+
+        try:
+            # have zipfile
+            z = zipfile.ZipFile(filecontent)
         except zipfile.BadZipfile:
-            self.logger.error('%s: unzip failure for content:\n%s' % (self.NAME, content))
+            LOGGER.error('%s: unzip failure for content:\n%s' % (self.NAME, content))
             return None
 
-        unzipped = z.read(z.namelist()[0]) # have unzipped content
+        # have unzipped content
+        unzipped = [z.read(thisfile) for thisfile in z.namelist()]
         z.close()
+
+        # return
         return unzipped
 
     def parse_to_df(self, filelike, mode='csv', header_names=None, sheet_names=None, **kwargs):
@@ -288,35 +322,56 @@ class BaseClient(object):
         if mode == 'csv':
             # convert string to filelike if needed
             try:
-                is_closed = filelike.closed
-            except AttributeError: # string, unicode, etc
-                filelike = StringIO(filelike)
+                filelike.closed
+            except AttributeError:  # string, unicode, etc
+                try:
+                    filelike = BytesIO(filelike)  # This was changed from StringIO for Python 3.x
+                except TypeError:
+                    filelike = StringIO(filelike)
 
             # read csv
             df = pd.read_csv(filelike, **kwargs)
 
         # do xls
         elif mode == 'xls':
+            # parse_dates is not implemented for excel, so pop it off
+            if 'parse_dates' in kwargs:
+                parse_dates = kwargs.pop('parse_dates')
+            else:
+                parse_dates = False
+
             pieces = []
             for sheet in sheet_names:
                 pieces.append(filelike.parse(sheet, **kwargs))
             df = pd.concat(pieces)
 
+            # parse date index
+            if parse_dates:
+                idx = []
+                for t in df.index:
+                    try:
+                        idx.append(pd.to_datetime(t))
+                    except ValueError:
+                        # not a datetime
+                        idx.append('')
+                df.index = idx
+
         # set names
         if header_names is not None:
             df.columns = header_names
 
         # drop na
         df = df.dropna()
-                
+
         return df
 
-    def utcify_index(self, local_index, tz_name=None):
+    def utcify_index(self, local_index, tz_name=None, tz_col=None):
         """
         Convert a DateTimeIndex to UTC.
 
         :param DateTimeIndex local_index: The local DateTimeIndex to be converted.
-        :param string tz_name: If local_ts is naive, it is assumed to be in timezone tz. If tz is not provided, the client's default timezone is used.
+        :param string tz_name: If local_ts is naive, it is assumed to be in timezone tz.
+            If tz is not provided, the client's default timezone is used.
         :return: DatetimeIndex in UTC.
         :rtype: DatetimeIndex
         """
@@ -324,15 +379,38 @@ class BaseClient(object):
         if tz_name is None:
             tz_name = self.TZ_NAME
 
-        # localize
-        try:
-            aware_local_index = local_index.tz_localize(tz_name)
-        except Exception as e:
-            self.logger.debug(e)  # already aware
-            aware_local_index = local_index
+        # use tz col if given
+        if tz_col is not None:
+            # it seems like we shouldn't have to iterate, but all the smart ways aren't working
+            aware_utc_list = []
+            for i in range(len(local_index)):
+                try:
+                    aware_local_ts = pytz.timezone(tz_col[i]).localize(local_index[i])
+                except pytz.UnknownTimeZoneError:
+                    # fall back to local ts
+                    aware_local_ts = pytz.timezone(tz_name).localize(local_index[i])
 
-        # convert to utc
-        aware_utc_index = aware_local_index.tz_convert('UTC')
+                # utcify
+                aware_utc_ts = self.utcify(aware_local_ts)
+                aware_utc_list.append(aware_utc_ts)
+
+            # indexify
+            aware_utc_index = pd.DatetimeIndex(aware_utc_list)
+
+        else:
+            # localize
+            try:
+                aware_local_index = local_index.tz_localize(tz_name)
+            except AmbiguousTimeError as e:
+                LOGGER.debug(e)
+                aware_local_index = local_index.tz_localize(tz_name, ambiguous='infer')
+            except TypeError as e:
+                # already aware
+                LOGGER.debug(e)
+                aware_local_index = local_index
+
+            # convert to utc
+            aware_utc_index = aware_local_index.tz_convert('UTC')
 
         # return
         return aware_utc_index
@@ -341,6 +419,10 @@ class BaseClient(object):
         if options is None:
             options = self.options
 
+        if len(df) == 0:
+            # if empty, end here
+            return df
+
         if options.get('latest', None):
             start_at = df.iloc[-1].name
             end_at = start_at
@@ -351,7 +433,12 @@ class BaseClient(object):
             except KeyError:
                 raise ValueError('Slicing by time requires start_at and end_at')
 
-        return df.truncate(before=start_at, after=end_at)
+        # sort before truncate eliminates DST KeyError
+        sorteddf = df.sort()
+        sliced = sorteddf.truncate(before=start_at, after=end_at)
+
+        # return
+        return sliced
 
     def unpivot(self, df):
         return df.stack().reset_index(level=1)
@@ -366,6 +453,18 @@ class BaseClient(object):
 
         return data
 
+    def serialize_faster(self, df, extras={}):
+        """DF is a DataFrame with DateTimeIndex and columns fuel_type and gen_MW (or load_mW).
+        Index and columns are already properly named."""
+        df = df.reset_index()
+        for key in extras:
+            df[key] = extras[key]
+        return df.to_dict(orient='records')
+
+    def local_now(self):
+        """Returns a tz-aware datetime equal to the current moment, in the local timezone"""
+        return pytz.utc.localize(datetime.utcnow()).astimezone(pytz.timezone(self.TZ_NAME))
+
     def dates(self):
         """Returns a list of dates in local time"""
         # set up storage
@@ -373,7 +472,7 @@ class BaseClient(object):
 
         # if latest, use date in local time
         if self.options['latest']:
-            local_now = pytz.utc.localize(datetime.utcnow()).astimezone(pytz.timezone(self.TZ_NAME))
+            local_now = self.local_now()
             if local_now.date() != (local_now - timedelta(minutes=30)).date():
                 dates.append((local_now - timedelta(minutes=30)).date())
             dates.append(local_now.date())
@@ -389,7 +488,8 @@ class BaseClient(object):
 
         # have to have some sort of dates
         else:
-            raise ValueError('Either latest must be True, or start_at and end_at must both be provided.')
+            raise ValueError(
+                'Either latest must be True, or start_at and end_at must both be provided.')
 
         # return
         return dates
