@@ -1,14 +1,15 @@
 from datetime import datetime, timedelta
 import pytz
+from dateutil.parser import parse as dateutil_parse
 import pandas as pd
 from pyiso.base import BaseClient
 
 
 class BPAClient(BaseClient):
     NAME = 'BPA'
-    
+
     base_url = 'http://transmission.bpa.gov/business/operations/'
-    
+
     fuels = {
         'Hydro': 'hydro',
         'Wind': 'wind',
@@ -16,7 +17,7 @@ class BPAClient(BaseClient):
     }
 
     TZ_NAME = 'America/Los_Angeles'
-       
+
     def fetch_historical(self):
         """Get BPA generation or load data from the far past"""
         # set up requests
@@ -45,11 +46,11 @@ class BPAClient(BaseClient):
         for url in request_urls:
             xd = self.fetch_xls(url)
             piece = self.parse_to_df(xd, mode='xls', sheet_names=xd.sheet_names,
-                                    skiprows=18, parse_cols=cols,
-                                    parse_dates=True, index_col=0,
-                                    header_names=header_names)
+                                     skiprows=18, parse_cols=cols,
+                                     index_col=0, parse_dates=True,
+                                     header_names=header_names)
             pieces.append(piece)
-            
+
         # return
         df = pd.concat(pieces)
         return df
@@ -70,10 +71,15 @@ class BPAClient(BaseClient):
 
         # parse like tsv
         df = self.parse_to_df(response.text, skiprows=6, header=0, delimiter='\t',
-                            index_col=0, parse_dates=True, usecols=cols)
+                              index_col=0, usecols=cols,
+                              date_parser=self.date_parser)
 
         return df
 
+    def date_parser(self, ts_str):
+        ts = dateutil_parse(ts_str)
+        return ts
+
     def fetcher(self):
         """Choose the correct fetcher method for this request"""
         # get mode from options
@@ -93,17 +99,17 @@ class BPAClient(BaseClient):
         return fetcher
 
     def parse_generation(self, df):
+        # process times
+        df.index = self.utcify_index(df.index)
+        sliced = self.slice_times(df)
+
         # original header is fuel names
-        df.rename(columns=self.fuels, inplace=True)
-        pivoted = self.unpivot(df)
+        sliced.rename(columns=self.fuels, inplace=True)
+        pivoted = self.unpivot(sliced)
         pivoted.rename(columns={'level_1': 'fuel_name', 0: 'gen_MW'}, inplace=True)
 
-        # process times
-        pivoted.index = self.utcify_index(pivoted.index)
-        sliced = self.slice_times(pivoted)
-
         # return
-        return sliced
+        return pivoted
 
     def handle_options(self, **kwargs):
         # default handler
