@@ -37,3 +37,19 @@ def get_load(ba_name, **kwargs):
     
     # return
     return data
+
+
+@shared_task
+def get_trade(ba_name, **kwargs):
+    # get data
+    c = client_factory(ba_name)
+    data = c.get_trade(**kwargs)
+    
+    # log
+    if len(data) == 0:
+        msg = '%s: No trade data at %s with args %s' % (ba_name, datetime.utcnow().isoformat(),
+                                                    kwargs)
+        logger.warn(msg)
+    
+    # return
+    return data
